
import numpy as np
import pandas as pd
import tensorflow as tf
import scipy.linalg as la

from rdkit import Chem as chem
from rdkit.Chem import rdmolops
from rdkit.Chem import MolFromSmiles as ms


class Conv(object):
    possible_atom_list = ['C', 'N', 'O', 'S', 'F', 'P', 'Cl', 'Mg',
                          'Na', 'Br', 'Ca', 'Cu', 'Pd', 'I', 'Al']
    possible_valence_list = list(range(9))
    possible_deg_list = list(range(9))
    possible_hybridization_list = [chem.rdchem.HybridizationType.SP,
                                   chem.rdchem.HybridizationType.SP2,
                                   chem.rdchem.HybridizationType.SP3,
                                   chem.rdchem.HybridizationType.SP3D,
                                   chem.rdchem.HybridizationType.SP3D2]

    def __init__(self, smiles):
        '''Initialize Conv object from SMILES string'''
        self.smiles = smiles
        mol = ms(smiles)

        # Create atom features
        raw_nodes = [self._get_atom_features(a) for a in mol.GetAtoms()]
        atom_features = np.vstack(raw_nodes)
        self.num_atoms, self.num_feat = atom_features.shape
        pad = 130 - self.num_atoms

        # Pad atom features
        feat_paddings = tf.constant([[0, pad], [0,0]])
        self.atom_features = tf.pad(tf.constant(atom_features, dtype=tf.float32),
                                    feat_paddings, 'CONSTANT').numpy()

        # Pad and normalize adjacency matrix
        adj_paddings = tf.constant([[0, pad], [0,pad]])
        A = rdmolops.GetAdjacencyMatrix(mol) + np.eye(mol.GetNumAtoms())
        D = la.inv(la.sqrtm(np.diag(np.sum(A, axis=1))))
        A = np.matmul(np.matmul(D, A), D)

        self.adj_matrix = tf.pad(tf.constant(A, dtype=tf.float32),
                                 adj_paddings, 'CONSTANT').numpy()

    def _get_one_hot_unk(self, x, allowable_set):
        '''NOTE: Maps inputs not in the allowable set to the last element'''
        if x not in allowable_set:
            x = allowable_set[-1]

        return list(map(lambda s: x == s, allowable_set))

    def _get_atom_features(self, atom):
        results = (self._get_one_hot_unk(atom.GetSymbol(), self.possible_atom_list) +
                   self._get_one_hot_unk(atom.GetDegree(), self.possible_deg_list) +
                   self._get_one_hot_unk(atom.GetImplicitValence(), self.possible_valence_list) +
                   [atom.GetFormalCharge(), atom.GetNumRadicalElectrons()] +
                   self._get_one_hot_unk(atom.GetHybridization(), self.possible_hybridization_list) +
                   [atom.GetIsAromatic()])

        return np.array(results)




import os

import numpy as np
import pandas as pd

from rdkit import Chem as chem
from rdkit.Chem import MolFromSmiles as ms


def read_data():
    '''
    Reads reaction data files in /rxn-data/ and concatenates them into one.
    Output: pandas DataFrame containing reaction SMARTS and atom-labeled bond
        changes from the US Patent Office database.
    '''
    current_dir = os.path.dirname(os.path.realpath('__file__'))

    train_data_file = os.path.join(current_dir, 'odachi/data/rxn-data/train.txt')
    valid_data_file = os.path.join(current_dir, 'odachi/data/rxn-data/valid.txt')
    test_data_file = os.path.join(current_dir, 'odachi/data/rxn-data/test.txt')

    train = pd.read_csv(train_data_file, sep=' ', engine='python', header=None)
    valid = pd.read_csv(valid_data_file, sep=' ', engine='python', header=None)
    test = pd.read_csv(test_data_file, sep=' ', engine='python', header=None)

    return train.append(valid, ignore_index=True).append(test, ignore_index=True)


def mol_with_atom_index(mol):
    '''
    Convert atom indexes in RDKit mol from given to canonicalized values.
    Input: RDKit mol with custom or no atom indexes.
    Output: RDKit mol with canonicalized atom indexes.
    '''
    num_atoms = mol.GetNumAtoms()

    for idx in range(num_atoms):
        mol.GetAtomWithIdx(idx).SetProp('molAtomMapNumber', str(mol.GetAtomWithIdx(idx).GetIdx()))

    return mol


def _rxn_to_mol(data):
    '''
    Parse reaction SMARTS for molecular SMILES of reactants and product for
    all entries in data.
    Input: DataFrame with reaction SMARTS in first column.
    Output: react: List of reactant SMILES
            prod: List of product SMILES
    '''
    react, prod = list(), list()

    for rxn in data[0]:
        r, p = rxn.split(">>")
        react.append(r)
        prod.append(p)

    return react, prod


def _bond_to_pair(data):
    '''
    Parse striings of changed bonds into tuples of bonds as pairs of atoms.
    Input: DataFrame with changed bond string in second column.
    Output: bond_list: List of list of changed bond tuples.
    '''
    bond_list = list()

    for bond_strings in data[1]:
        bonds = list()
        bond_str = bond_strings.split(';')

        for bond_str in bond_strings.split(';'):
            bond = tuple(map(int, bond_str.split('-')))
            bonds.append(bond)

        bond_list.append(bonds)

    return bond_list


def process_data(data):
    '''
    Combine operations of rxn_to_mol and bond_to_pair to create a new DataFrame.
    Input: Initial DataFrame.
    Output: Reprocessed DataFrame with changed bond tuples and separatated
        reactant and product SMILES.
    '''
    react, prod = _rxn_to_mol(data)
    bonds = _bond_to_pair(data)

    return pd.DataFrame(zip(*[bonds, prod, react]), columns=["Bonds", "Products", "Reactants"])


def _find_in_smiles(idx, reactants):
    '''
    Find atom label in list of reactant SMILES.
    Input:  idx: atom index.
            reactants: list of reactant SMILES.
    Output: reactant: SMILES of reactant that has the indicated atom.
    '''
    search = f':{idx}]'
    for reactant in reactants:
        if search in reactant:
            return reactant


def assign_clusters(idx, data):
    '''
    Takes as an input a data entry, and outputs a list with labels of what
    atoms in the product come from which reactants.
    Input:  idx: Entry index.
            data: Dataset after process_data operation.
    Output: mapping: List of correspondances between product atoms in reactants.
    '''
    reacts = data["Reactants"][idx].split('.')
    mol = ms(data["Products"][idx])
    atoms = mol.GetNumAtoms()
    mapping = list()

    for i in range(atoms):
        atom_str = mol.GetAtomWithIdx(i).GetProp('molAtomMapNumber')
        mapping.append(_find_in_smiles(atom_str, reacts))
        mol.GetAtomWithIdx(i).SetProp('molAtomMapNumber', str(mol.GetAtomWithIdx(i).GetIdx()))

    distinct = list(sorted(set(mapping)))

    for j in range(atoms):
        mapping[j] = distinct.index(mapping[j])

    return mapping




import os
import glob
import itertools
import numpy as np
import tensorflow as tf

from collections import deque
from odachi.engine.layers import ConvEmbed, conv_classifier


def map2retro(mapping):
    '''Convert mapping list into retrosynthetic similarity matrix.'''
    ret = np.logical_xor.outer(mapping, mapping)
    return np.logical_not(ret).astype(np.float32)


def train_batch(file):
    '''Use subbatch of Conv molecule to train the model on subset of possible atom combinations.'''
    loss = 0

    with tf.GradientTape() as tape:
        # Define subbatch size
        for _ in range(100):

            # Open Conv object:
            conv = pickle.load(file)

            # Perform DGCN embedding
            embed = odachi_embed([conv.adj_matrix, conv.atom_features])

            # Define possible indexes and targets
            idxs = list(itertools.combinations(np.arange(conv.num_atoms), 2))
            retro = map2retro(conv.mapping)[np.triu_indices(conv.num_atoms, 1)]
            targs = tf.keras.utils.to_categorical(retro, 2)
            vecs = deque(maxlen=len(idxs))

            # Classify retrosynthetic similarity on all possible combinations of feature vectors
            for aidx in idxs:
                vecs.append(tf.reshape(tf.gather(embed, aidx, axis=1), [1, 82]))

            vec = tf.concat(list(vecs), 0)
            out = classifier(vec)
            loss += loss_fn(targs, out)

    # Define model stack variables
    variables = odachi_embed.trainable_variables + \
                classifier.trainable_variables

    # Calculate and apply gradients
    grads = tape.gradient(loss, variables)
    optimizer.apply_gradients(zip(grads, variables))

    return loss


def train_file(filename, idx=None):
    '''Train model stack on all subbatches of Conv objects in a file - defines one 'batch'.'''
    loss_avg = tf.keras.metrics.Mean()
    file = open(filename, 'rb')

    for i in range(10):
        batch_loss = train_batch(file)
        loss_avg(batch_loss)
    file.close()
    # print(f'Epoch {epoch} Batch {idx} Loss {loss_avg.result().numpy():.10f} Time {time.time() - start:.4f}')

    return loss_avg.result().numpy()


def validate_single(conv, tries=10):
    '''Validate model performance on a single Conv.'''
    metrics = tf.keras.metrics.BinaryAccuracy()
    embed = odachi_embed([conv.adj_matrix, conv.atom_features])
    idxs = list(itertools.combinations(np.arange(conv.num_atoms), 2))
    tries = len(idxs) if len(idxs) < tries else tries

    for aidx in random.sample(idxs, tries):
        targ = tf.constant([[0, 1]], dtype=tf.float32) \
               if conv.mapping[aidx[0]] == conv.mapping[aidx[1]] \
               else tf.constant([[1, 0]], dtype=tf.float32)

        vec = odachi_ext([embed, aidx])
        out = classifier(vec)
        metrics.update_state(targ, out)

    return metrics.result().numpy()


def validate(filename, tries=10):
    '''Validate model performance on all Conv objects in a file.'''
    # print('\nValidating performance')
    val_acc_avg = tf.keras.metrics.Mean()
    file = open(filename, 'rb')

    while True:
        try:
            conv = pickle.load(file)

        except EOFError:
            break

        acc = validate_single(conv, tries)
        val_acc_avg(acc)

    file.close()
    # print(f'Epoch {epoch} Accuracy {val_acc_avg.result().numpy():.6f}')

    return val_acc_avg.result().numpy()


def train_epoch():
    '''Train model for a single epoch: iterate through all Conv files.'''
    convs = glob.glob('data/*')
    epoch_loss = tf.keras.metrics.Mean()
    valid_acc = tf.keras.metrics.Mean()

    for idx, convfile in enumerate(convs[:-5]):
        batch_loss = train_file(convfile, idx)
        epoch_loss(batch_loss)

    for convfile in convs[-5:]:
        val_acc = validate(convs[-1])
        valid_acc(val_acc)

    return epoch_loss.result().numpy(), valid_acc.result().numpy()


def save_model(idx):
    '''Save models for a given epoch number.'''
    odachi_embed.save_weights(f'odachi_embed_v1_e{idx}.h5')
    classifier.save(f'odachi_class_v1_e{idx}.h5')


if __name__ == '__main__':
    convs = glob.glob(os.path.join(os.path.dirname(os.path.realpath('__file__')), '/conv-data/*'))
    odachi_embed, classifier = ConvEmbed(4, 0.05), conv_classifier()

    optimizer = tf.keras.optimizers.Adam(learning_rate=1e-3)
    loss_fn = tf.keras.losses.CategoricalCrossentropy()

    EPOCHS = 10

    for epoch in range(EPOCHS):
        print(f'Beginning epoch {epoch}')

        epoch_loss, val_accs = train_epoch()
        print(f'Epoch {epoch} Training Loss {epoch_loss:.10f}')
        print(f'Validation Accuracy {val_accs:.4f}')
        save_model(epoch)




import os
import time
import itertools
import numpy as np
import tensorflow as tf

from collections import deque
from sklearn.cluster import KMeans

from rdkit import Chem as chem
from rdkit.Chem import rdDepictor, rdmolops
from rdkit.Chem.Draw import rdMolDraw2D
from rdkit.Chem import MolFromSmiles as ms

from odachi.data.read import mol_with_atom_index
from odachi.data.conv import Conv
from odachi.engine.layers import ConvEmbed

MODEL_DIR = os.path.dirname(os.path.abspath(__file__))


class OdachiEngine(tf.keras.Model):
    def __init__(self, version=9, knock=0):
        super(OdachiEngine, self).__init__()

        self.embed = ConvEmbed(4, knock)
        self.embed.load_weights(os.path.join(MODEL_DIR, f'models/odachi_embed_v1_e{version}.h5'))

        self.classifier = tf.keras.models.load_model(os.path.join(MODEL_DIR, f'models/odachi_class_v1_e{version}.h5'), compile=False)

    def call(self, adj_matrix, atom_features, num_atoms):
        embed = self.embed([adj_matrix, atom_features])

        idxs = list(itertools.combinations(np.arange(num_atoms), 2))
        vecs = deque(maxlen=len(idxs))

        for idx in idxs:
            vecs.append(tf.reshape(tf.gather(embed, idx, axis=1), [1, 82]))

        vec = tf.concat(list(vecs), 0)

        return idxs, self.classifier(vec)


class Odachi:
    def __init__(self, knock=0):
        self.engines = [OdachiEngine(i, knock) for i in range(10)]

    def _build_adj(self, conv, idxs, classes):
        similarity = np.eye(conv.num_atoms)

        for (i, j), cl in zip(idxs, classes):
            similarity[i, j] = similarity[j, i] = cl[-1]

        return similarity

    def _cluster(self, similarity, num_clusters=2):
        D = np.diag(similarity.sum(axis=1))
        L = D - similarity

        vals, vecs = np.linalg.eig(L)
        vecs = vecs[:, np.argsort(vals)]
        vals = vals[np.argsort(vals)]

        kmeans = KMeans(n_clusters=num_clusters)
        kmeans.fit(vecs[:,1:num_clusters])
        colors = kmeans.labels_

        return colors

    def _get_broken_bonds(self, mol, colors):
        bonds = []

        for idx in range(mol.GetNumBonds()):
            a = mol.GetBondWithIdx(idx).GetBeginAtomIdx()
            b = mol.GetBondWithIdx(idx).GetEndAtomIdx()
            if colors[a] != colors[b]:
                bonds.append(idx)

        return bonds

    def __call__(self, smiles, clusters, version):
        assert 0 <= version <= 9, "Version error: pick version between 0 and 9"

        start = time.time()

        imol, iconv = ms(smiles), Conv(smiles)

        similarity = self._build_adj(iconv, *self.engines[version](iconv.adj_matrix, iconv.atom_features, iconv.num_atoms))
        colors = self._cluster(similarity, clusters)
        bonds = self._get_broken_bonds(imol, colors)

        rdDepictor.Compute2DCoords(imol)
        drawer = rdMolDraw2D.MolDraw2DSVG(600,300)
        drawer.DrawMolecule(imol, highlightAtoms=[], highlightBonds=bonds)
        drawer.FinishDrawing()

        svg = drawer.GetDrawingText().replace('svg:','')
        # open(os.path.join(os.path.dirname(os.path.realpath('__file__')), f'{smiles}.svg'), 'w').write(svg)

        return {'smiles': smiles, 'svg': svg, 'bonds': bonds, 'time': time.time() - start}




import os
import argparse
import numpy as np
import tensorflow as tf

from odachi.engine.model import Odachi


def parse():
    parser = argparse.ArgumentParser(description='Run Odachi Retrosynthesis Engine on smiles string')
    parser.add_argument('smiles', type=str, help='a SMILES string for the molecule')
    parser.add_argument('--clusters', type=int, default=2,
                        help='number of disconnections')

    return parser


if __name__ == '__main__':
    odachi = Odachi()
    parser = parse()
    args = parser.parse_args()

    odachi(args.smiles, args.clusters)




import numpy as np
import tensorflow as tf


class GraphConv(tf.keras.layers.Layer):
    '''Tensorflow layer for performing Kipf-style graph convolutions'''
    def __init__(self, n,
                 num_feat=41,
                 num_atoms=130,
                 activation=tf.nn.elu,
                 dropout = 0.1,
                 BATCH_SIZE=1):

        super(GraphConv, self).__init__(autocast=False)

        self.n = n
        self.num_feat = num_feat
        self.num_atoms = num_atoms
        self.activation = activation
        self.batch_size = BATCH_SIZE

        w_init = tf.random_normal_initializer()
        b_init = tf.zeros_initializer()

        self.w = tf.Variable(initial_value=w_init(shape=(self.batch_size,
                                                         self.num_feat,
                                                         self.num_feat),
                                                  dtype='float32'),
                             trainable=True,
                             name=self.n + '_w')

        self.b = tf.Variable(initial_value=b_init(shape=(self.batch_size,
                                                         self.num_atoms,
                                                         self.num_feat),
                                                  dtype='float32'),
                             trainable=True,
                             name=self.n + '_b')

        self.drop = lambda : tf.cast(tf.greater_equal(tf.random.uniform((self.batch_size,
                                                                         self.num_feat,
                                                                         self.num_feat)),
                                                      dropout),
                                     tf.float32)

    @tf.function
    def call(self, inputs):
        A, X = inputs
        X = self.activation(A @ (X @ (self.drop() * self.w)) + self.b)
        return A, X


class ConvEmbed(tf.keras.Model):
    '''Convenience class of stacked convolutional layers'''
    def __init__(self,
                 depth = 10,
                 drop = 0.2,
                 BATCH_SIZE=1):
        super(ConvEmbed, self).__init__()

        self.layer_stack = []

        for idx in range(depth):
            n = 'conv_' + str(idx)
            self.layer_stack.append(GraphConv(n, dropout=drop, BATCH_SIZE=BATCH_SIZE))

    def call(self, I):
        for layer in self.layer_stack:
            I = layer(I)
        return I[1]


class ExtractVec(tf.keras.layers.Layer):
    '''Pulls comparison vector from convolved features matrix'''
    def __init__(self, feat=41):
        super(ExtractVec, self).__init__()
        self.feat = feat

    def call(self, inputs):
        X, (ax, ay) = inputs

        a = np.zeros((1, 130, 1))
        a[:, ax], a[:, ay] = 1, 1
        an = tf.squeeze(tf.cast(a, tf.bool), axis=0)
        an = tf.broadcast_to(an, [130, self.feat])

        X.set_shape((1, 130, self.feat))
        X = tf.boolean_mask(X, an, axis=1)
        X.set_shape((1, 2 * self.feat))

        return X


def conv_classifier():
    classifier = tf.keras.models.Sequential(name='odachi_classifier')
    classifier.add(tf.keras.layers.Dropout(0.5, input_shape=(82,)))

    classifier.add(tf.keras.layers.Dense(60, activation = tf.nn.elu))
    classifier.add(tf.keras.layers.Dropout(0.5))

    classifier.add(tf.keras.layers.Dense(40, activation = tf.nn.elu))
    classifier.add(tf.keras.layers.Dropout(0.5))

    classifier.add(tf.keras.layers.Dense(30, activation = tf.nn.elu))
    classifier.add(tf.keras.layers.Dropout(0.5))

    classifier.add(tf.keras.layers.Dense(20, activation = tf.nn.elu))
    classifier.add(tf.keras.layers.Dropout(0.5))

    classifier.add(tf.keras.layers.Dense(2, activation = 'softmax'))

    return classifier




import os
import uuid
import pickle

import numpy as np
import pandas as pd
import tensorflow as tf
import scipy.linalg as la

from rdkit import Chem as chem
from rdkit.Chem import MolFromSmiles as ms

import odachi.data


def write_batch(start, name, data, batch_size=1000):
    '''Write batch of data to a pickle file of Conv objects.'''
    outfile = open(name, 'wb+')

    for idx in range(start, start+batch_size):
        conv = odachi.data.conv.Conv(data.Products[idx])
        conv.mapping = odachi.data.read.assign_clusters(idx, data)
        pickle.dump(conv, outfile)

    outfile.close()


def write_files():
    '''Converts all molecules in data to Conv objects for the model to train on.'''
    raw_data = odachi.data.read.read_data()
    data = odachi.data.read.process_data(raw_data)

    path = os.path.join(os.path.dirname(os.path.realpath('__file__')), '/conv-data/')
    os.mkdir(path)

    for i in range(data.shape[0] // 1000):
        write_batch(i * 1000, path + uuid.uuid4().hex, data)




from setuptools import setup
from os import path

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
  name = 'odachi',         # How you named your package folder (MyLib)
  packages = ['odachi'],   # Chose the same as "name"
  version = '0.2.5',      # Start with a small number and increase it with every change you make
  license='MIT',        # Chose a license from here: https://help.github.com/articles/licensing-a-repository
  description = 'Advanced deep learning-based organic retrosynthesis engine',   # Give a short description about your library
  long_description = long_description,
  long_description_content_type='text/markdown',
  author = 'Rushil Mallarapu',                   # Type in your name
  author_email = 'rushil.mallarapu@gmail.com',      # Type in your E-Mail
  url = 'https://github.com/sudo-rushil/odachi',   # Provide either the link to your github or to your website
  download_url = 'https://github.com/sudo-rushil/odachi/archive/v0.2.tar.gz',    # I explain this later on
  keywords = ['Chemistry', 'Organic Chemistry', 'Synthesis', 'Retrosynthesis', 'Deep Learning', 'AI', 'Graph', 'Convolution', 'Graph Convolution', 'CNN'],
  include_package_data=True,
  install_requires=[            # I get to this in a second
          'numpy',
          'pandas',
          'scikit-learn',
          'tensorflow',
      ],
  classifiers=[
    'Development Status :: 4 - Beta',      # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
    'Intended Audience :: Developers',      # Define that your audience are developers
    'Topic :: Software Development :: Build Tools',
    'License :: OSI Approved :: MIT License',
    'Programming Language :: Python :: 3.7',
  ])
